unit FastList;
{
  FastList class library
  Delphi Fast Search Strings

   Zinkevich Viktor, 2004

  If You use this module for comercial purpose
  You would write information in a special information
  box or documentation:

  Partial Copyright  2004 Zinkevich Viktor
  http://vicking.narod.ru/
  mailto:viktor-zin@mail.ru

}


interface

uses SysUtils, Classes;

type
  TCond = record
    Con: Byte;
    Res: Integer;
  end;

  TState = class(TObject)
  private
    Table: array of TCond;
  protected

  public
    function Switch(Cond: Byte): Integer;
    function AddCond(Cond: TCond): Integer;
    destructor Destroy; override;
  published

  end;

  TFastList = class(TObject)
  private
    List: TList;
    FCount: Integer;
    FStrings: TStrings;
    property Count: Integer read FCount;
    function Add(S: string): Integer;

  protected

  public
    constructor Create;
    procedure Clear;
    destructor Destroy; override;
    function IndexOf(S: string): Integer;
    property Strings: TStrings read FStrings;
    procedure Compile;
  published

  end;


implementation

{ TState }

// Add a new condition to the State
function TState.AddCond(Cond: TCond): Integer;
var
  i: Integer;
begin
  // Is there such condition in this state?
  i := Switch(Cond.Con);
  if i = -1 then
  begin
    // Add a new condition
    SetLength(Table,Length(Table)+1);
    Table[High(Table)].Con := Cond.Con;
    Table[High(Table)].Res := Cond.Res;
    Result := High(Table); // Return its index
  end
  else
    Result := -i; // return negative index
end;
// destructor
destructor TState.Destroy;
begin
  // Free up the memory
  Finalize(Table);
  inherited;
end;

// Switches into new state under specified condition
function TState.Switch(Cond: Byte): Integer;
var
  i: Byte;
begin
  Result := -1;
  if Length(Table)=0 then Exit;
  i := High(Table);
  while (i>0) and (Table[i].Con <> Cond) do
    Dec(i);
  if (Table[i].Con <> Cond) then Exit;
  Result:= Table[i].Res;
end;

{ TFastList }

// Add a new string into compiled list
function TFastList.Add(S: string): Integer;
var
  pc: PChar;
  i, n: Integer;
  st: TState;
  Cond: TCond;
begin
  pc := PChar(S);
  st := TState(List.First);

  for i:=0 to Length(S) do
  begin

    Cond.Con := Byte(pc[i]);
    Cond.Res := List.Count;
    if Cond.Con = 0 then Cond.Res := Count;
    n := st.AddCond(Cond);
    if n < 0 then
    begin
      st := TState(List.Items[-n]);
    end
    else
    begin
      st := TState.Create;
      List.Add(st);
    end;
  end;
  Result := Count;
  Inc(FCount);
end;

// Clear compiled list
procedure TFastList.Clear;
var
  i: Integer;
begin
  for i:= 0 to List.Count-1 do
  begin
    TState(List.Items[i]).Free;
  end;
  List.Clear;
  // one state must always be created
  List.Add(TState.Create);
end;

// Compile Strings into special codes
procedure TFastList.Compile;
var
  i: Integer;
begin
  Clear;
  for i:=0 to Strings.Count-1 do
  begin
    Add(Strings.Strings[i]);
  end;
end;

// Constructor
constructor TFastList.Create;
begin
  List := TList.Create;
  List.Add(TState.Create);
  FStrings := TStringList.Create;
end;

// Destructor
destructor TFastList.Destroy;
begin
  Clear;
  List.Free;
  FStrings.Free;
  inherited;
end;


// Return index of specified string or -1
function TFastList.IndexOf(S: string): Integer;
var
  pc: PChar;
  i, n: Integer;
  st: TState;
  b: Byte;
begin
  pc := PChar(S);
  st := TState(List.First);
  Result := -1;
  for i:=0 to Length(S) do
  begin
    b := Byte(pc[i]);
    n := st.Switch(b);
    if b = 0 then Result := n
    else
      if n <> -1 then
      begin
        st := TState(List.Items[n]);
      end
      else
        Result := -1;
  end;
end;


end.
